@@ -33,11 +33,6 @@
 
 
 
-#ifndef MAX
-#  define MIN(a,b)      ((a<b) ? a : b)
-#  define MAX(a,b)      ((a>b) ? a : b)
-#endif
-
 #define SKIP_MASK(x,y) (mask_rb && !*(mask_rb->data.rgba.a+(y)*mask_rb->data.rgba.row_step+(x)*mask_rb->data.rgba.pixel_step))
 
 /************************************************************************/
@@ -99,9 +94,7 @@ msNearestRasterResampler( imageObj *psSrcImage, rasterBufferObj *src_rb,
   int   nSrcXSize = psSrcImage->width;
   int   nSrcYSize = psSrcImage->height;
   int   nFailedPoints = 0, nSetPoints = 0;
-#ifndef USE_GD
-  assert(!MS_RENDERER_PLUGIN(psSrcImage->format) || src_rb->type != MS_BUFFER_GD);
-#endif
+  assert(!MS_RENDERER_PLUGIN(psSrcImage->format) || src_rb->type == MS_BUFFER_BYTE_RGBA);
 
 
   x = (double *) msSmallMalloc( sizeof(double) * nDstXSize );
@@ -141,60 +134,47 @@ msNearestRasterResampler( imageObj *psSrcImage, rasterBufferObj *src_rb,
       }
 
       if( MS_RENDERER_PLUGIN(psSrcImage->format) ) {
-#ifdef USE_GD
-        if(src_rb->type == MS_BUFFER_GD) {
-          int   nValue = 0;
-          assert(!gdImageTrueColor(src_rb->data.gd_img));
-          nValue = panCMap[src_rb->data.gd_img->pixels[nSrcY][nSrcX]];
+        int src_rb_off;
+        rgbaArrayObj *src,*dst;
+        assert( src_rb->type == MS_BUFFER_BYTE_RGBA );
+        src = &src_rb->data.rgba;
+        dst = &dst_rb->data.rgba;
+        assert( src_rb && dst_rb );
+        src_rb_off = nSrcX * src->pixel_step
+                      + nSrcY * src->row_step;
 
-          if( nValue == -1 )
-            continue;
+        if( src->a == NULL || src->a[src_rb_off] == 255 ) {
+          int dst_rb_off;
+
+          dst_rb_off = nDstX * dst->pixel_step
+                        + nDstY * dst->row_step;
 
           nSetPoints++;
-          dst_rb->data.gd_img->pixels[nDstY][nDstX] = nValue;
 
-        } else
-#endif
-          if( src_rb->type == MS_BUFFER_BYTE_RGBA ) {
-            int src_rb_off;
-            rgbaArrayObj *src = &src_rb->data.rgba, *dst = &dst_rb->data.rgba;
-            assert( src_rb && dst_rb );
-            src_rb_off = nSrcX * src->pixel_step
-                         + nSrcY * src->row_step;
+          dst->r[dst_rb_off] = src->r[src_rb_off];
+          dst->g[dst_rb_off] = src->g[src_rb_off];
+          dst->b[dst_rb_off] = src->b[src_rb_off];
+          if( dst->a )
+            dst->a[dst_rb_off] = 255;
+        } else if( src->a[src_rb_off] != 0 ) {
+          int dst_rb_off;
 
-            if( src->a == NULL || src->a[src_rb_off] == 255 ) {
-              int dst_rb_off;
+          dst_rb_off = nDstX * dst->pixel_step
+                        + nDstY * dst->row_step;
 
-              dst_rb_off = nDstX * dst->pixel_step
-                           + nDstY * dst->row_step;
+          nSetPoints++;
 
-              nSetPoints++;
+          /* actual alpha blending is required */
+          msAlphaBlendPM( src->r[src_rb_off],
+                          src->g[src_rb_off],
+                          src->b[src_rb_off],
+                          src->a[src_rb_off],
+                          dst->r + dst_rb_off,
+                          dst->g + dst_rb_off,
+                          dst->b + dst_rb_off,
+                          dst->a ? dst->a + dst_rb_off : NULL  );
 
-              dst->r[dst_rb_off] = src->r[src_rb_off];
-              dst->g[dst_rb_off] = src->g[src_rb_off];
-              dst->b[dst_rb_off] = src->b[src_rb_off];
-              if( dst->a )
-                dst->a[dst_rb_off] = 255;
-            } else if( src->a[src_rb_off] != 0 ) {
-              int dst_rb_off;
-
-              dst_rb_off = nDstX * dst->pixel_step
-                           + nDstY * dst->row_step;
-
-              nSetPoints++;
-
-              /* actual alpha blending is required */
-              msAlphaBlendPM( src->r[src_rb_off],
-                              src->g[src_rb_off],
-                              src->b[src_rb_off],
-                              src->a[src_rb_off],
-                              dst->r + dst_rb_off,
-                              dst->g + dst_rb_off,
-                              dst->b + dst_rb_off,
-                              dst->a ? dst->a + dst_rb_off : NULL  );
-
-            }
-          }
+        }
       } else if( MS_RENDERER_RAWDATA(psSrcImage->format) ) {
         int band, src_off, dst_off;
 
@@ -259,16 +239,7 @@ static void msSourceSample( imageObj *psSrcImage, rasterBufferObj *rb,
   if( MS_RENDERER_PLUGIN(psSrcImage->format) ) {
     rgbaArrayObj *rgba;
     int rb_off;
-    assert(rb);
-#ifdef USE_GD
-    if(rb->type == MS_BUFFER_GD) {
-      assert(!gdImageTrueColor(rb->data.gd_img) );
-      padfPixelSum[0] += (dfWeight * rb->data.gd_img->pixels[iSrcY][iSrcX]);
-      *pdfWeightSum += dfWeight;
-      return;
-    }
-#endif
-    assert(rb->type == MS_BUFFER_BYTE_RGBA);
+    assert(rb && rb->type == MS_BUFFER_BYTE_RGBA);
     rgba = &(rb->data.rgba);
     rb_off = iSrcX * rgba->pixel_step + iSrcY * rgba->row_step;
 
@@ -343,7 +314,7 @@ msBilinearRasterResampler( imageObj *psSrcImage, rasterBufferObj *src_rb,
   int   nSrcYSize = psSrcImage->height;
   int   nFailedPoints = 0, nSetPoints = 0;
   double     *padfPixelSum;
-  int         bandCount = MAX(4,psSrcImage->format->bands);
+  int         bandCount = MS_MAX(4,psSrcImage->format->bands);
 
   padfPixelSum = (double *) msSmallMalloc(sizeof(double) * bandCount);
 
@@ -393,10 +364,10 @@ msBilinearRasterResampler( imageObj *psSrcImage, rasterBufferObj *src_rb,
         continue;
 
       /* Trim in stuff one pixel off the edge */
-      nSrcX = MAX(nSrcX,0);
-      nSrcY = MAX(nSrcY,0);
-      nSrcX2 = MIN(nSrcX2,nSrcXSize-1);
-      nSrcY2 = MIN(nSrcY2,nSrcYSize-1);
+      nSrcX = MS_MAX(nSrcX,0);
+      nSrcY = MS_MAX(nSrcY,0);
+      nSrcX2 = MS_MIN(nSrcX2,nSrcXSize-1);
+      nSrcY2 = MS_MIN(nSrcY2,nSrcYSize-1);
 
       memset( padfPixelSum, 0, sizeof(double) * bandCount);
 
@@ -419,54 +390,36 @@ msBilinearRasterResampler( imageObj *psSrcImage, rasterBufferObj *src_rb,
       if( dfWeightSum == 0.0 )
         continue;
 
-      for( i = 0; i < bandCount; i++ )
-        padfPixelSum[i] /= dfWeightSum;
-
       if( MS_RENDERER_PLUGIN(psSrcImage->format) ) {
         assert(src_rb && dst_rb);
-#ifdef USE_GD
-        if(src_rb->type == MS_BUFFER_GD) {
-          int nResult;
-          assert( !gdImageTrueColor(src_rb->data.gd_img) &&  !gdImageTrueColor(dst_rb->data.gd_img));
-          nResult = panCMap[(int) padfPixelSum[0]];
-          if( nResult != -1 ) {
-            nSetPoints++;
-            dst_rb->data.gd_img->pixels[nDstY][nDstX] = nResult;
-          }
-        } else
-#endif
-#ifdef USE_GD
-          if( src_rb->type == MS_BUFFER_BYTE_RGBA ) {
-#else
-          assert( src_rb->type == MS_BUFFER_BYTE_RGBA );
-#endif
-            assert( src_rb->type == dst_rb->type );
+        assert( src_rb->type == MS_BUFFER_BYTE_RGBA );
+        assert( src_rb->type == dst_rb->type );
 
 
-            nSetPoints++;
+        nSetPoints++;
 
-            if( dfWeightSum > 0.001 ) {
-              int dst_rb_off = nDstX * dst_rb->data.rgba.pixel_step + nDstY * dst_rb->data.rgba.row_step;
-              unsigned char red, green, blue, alpha;
+        if( dfWeightSum > 0.001 ) {
+          int dst_rb_off = nDstX * dst_rb->data.rgba.pixel_step + nDstY * dst_rb->data.rgba.row_step;
+          unsigned char red, green, blue, alpha;
 
-              red   = (unsigned char) MAX(0,MIN(255,padfPixelSum[0]));
-              green = (unsigned char) MAX(0,MIN(255,padfPixelSum[1]));
-              blue  = (unsigned char) MAX(0,MIN(255,padfPixelSum[2]));
-              alpha = (unsigned char)MAX(0,MIN(255,255.5*dfWeightSum));
+          red   = (unsigned char) MS_MAX(0,MS_MIN(255,padfPixelSum[0]));
+          green = (unsigned char) MS_MAX(0,MS_MIN(255,padfPixelSum[1]));
+          blue  = (unsigned char) MS_MAX(0,MS_MIN(255,padfPixelSum[2]));
+          alpha = (unsigned char)MS_MAX(0,MS_MIN(255,255.5*dfWeightSum));
 
-              msAlphaBlendPM( red, green, blue, alpha,
-                              dst_rb->data.rgba.r + dst_rb_off,
-                              dst_rb->data.rgba.g + dst_rb_off,
-                              dst_rb->data.rgba.b + dst_rb_off,
-                              (dst_rb->data.rgba.a == NULL) ? NULL : dst_rb->data.rgba.a + dst_rb_off );
-            }
-#ifdef USE_GD
-          }
-#endif
+          msAlphaBlendPM( red, green, blue, alpha,
+                          dst_rb->data.rgba.r + dst_rb_off,
+                          dst_rb->data.rgba.g + dst_rb_off,
+                          dst_rb->data.rgba.b + dst_rb_off,
+                          (dst_rb->data.rgba.a == NULL) ? NULL : dst_rb->data.rgba.a + dst_rb_off );
+        }
       } else if( MS_RENDERER_RAWDATA(psSrcImage->format) ) {
         int band;
         int dst_off = nDstX + nDstY * psDstImage->width;
 
+        for( i = 0; i < bandCount; i++ )
+            padfPixelSum[i] /= dfWeightSum;
+
         MS_SET_BIT(psDstImage->img_mask,dst_off);
 
         for( band = 0; band < psSrcImage->format->bands; band++ ) {
@@ -478,7 +431,7 @@ msBilinearRasterResampler( imageObj *psSrcImage, rasterBufferObj *src_rb,
               = (float) padfPixelSum[band];
           } else if( psSrcImage->format->imagemode == MS_IMAGEMODE_BYTE ) {
             psDstImage->img.raw_byte[dst_off]
-              = (unsigned char)MAX(0,MIN(255,padfPixelSum[band]));
+              = (unsigned char)MS_MAX(0,MS_MIN(255,padfPixelSum[band]));
           }
 
           dst_off += psDstImage->width*psDstImage->height;
@@ -531,14 +484,14 @@ msAverageSample( imageObj *psSrcImage, rasterBufferObj *src_rb,
   for( iY = nYMin; iY < nYMax; iY++ ) {
     double dfYCellMin, dfYCellMax;
 
-    dfYCellMin = MAX(iY,dfYMin);
-    dfYCellMax = MIN(iY+1,dfYMax);
+    dfYCellMin = MS_MAX(iY,dfYMin);
+    dfYCellMax = MS_MIN(iY+1,dfYMax);
 
     for( iX = nXMin; iX < nXMax; iX++ ) {
       double dfXCellMin, dfXCellMax, dfWeight;
 
-      dfXCellMin = MAX(iX,dfXMin);
-      dfXCellMax = MIN(iX+1,dfXMax);
+      dfXCellMin = MS_MAX(iX,dfXMin);
+      dfXCellMax = MS_MIN(iX+1,dfXMax);
 
       dfWeight = (dfXCellMax-dfXCellMin) * (dfYCellMax-dfYCellMin);
 
@@ -579,7 +532,7 @@ msAverageRasterResampler( imageObj *psSrcImage, rasterBufferObj *src_rb,
   int   nFailedPoints = 0, nSetPoints = 0;
   double     *padfPixelSum;
 
-  int         bandCount = MAX(4,psSrcImage->format->bands);
+  int         bandCount = MS_MAX(4,psSrcImage->format->bands);
 
   padfPixelSum = (double *) msSmallMalloc(sizeof(double) * bandCount);
 
@@ -615,19 +568,19 @@ msAverageRasterResampler( imageObj *psSrcImage, rasterBufferObj *src_rb,
         continue;
       }
 
-      dfXMin = MIN(MIN(x1[nDstX],x1[nDstX+1]),
-                   MIN(x2[nDstX],x2[nDstX+1]));
-      dfYMin = MIN(MIN(y1[nDstX],y1[nDstX+1]),
-                   MIN(y2[nDstX],y2[nDstX+1]));
-      dfXMax = MAX(MAX(x1[nDstX],x1[nDstX+1]),
-                   MAX(x2[nDstX],x2[nDstX+1]));
-      dfYMax = MAX(MAX(y1[nDstX],y1[nDstX+1]),
-                   MAX(y2[nDstX],y2[nDstX+1]));
+      dfXMin = MS_MIN(MS_MIN(x1[nDstX],x1[nDstX+1]),
+                   MS_MIN(x2[nDstX],x2[nDstX+1]));
+      dfYMin = MS_MIN(MS_MIN(y1[nDstX],y1[nDstX+1]),
+                   MS_MIN(y2[nDstX],y2[nDstX+1]));
+      dfXMax = MS_MAX(MS_MAX(x1[nDstX],x1[nDstX+1]),
+                   MS_MAX(x2[nDstX],x2[nDstX+1]));
+      dfYMax = MS_MAX(MS_MAX(y1[nDstX],y1[nDstX+1]),
+                   MS_MAX(y2[nDstX],y2[nDstX+1]));
 
-      dfXMin = MIN(MAX(dfXMin,0),psSrcImage->width+1);
-      dfYMin = MIN(MAX(dfYMin,0),psSrcImage->height+1);
-      dfXMax = MIN(MAX(-1,dfXMax),psSrcImage->width);
-      dfYMax = MIN(MAX(-1,dfYMax),psSrcImage->height);
+      dfXMin = MS_MIN(MS_MAX(dfXMin,0),psSrcImage->width+1);
+      dfYMin = MS_MIN(MS_MAX(dfYMin,0),psSrcImage->height+1);
+      dfXMax = MS_MIN(MS_MAX(-1,dfXMax),psSrcImage->width);
+      dfYMax = MS_MIN(MS_MAX(-1,dfYMax),psSrcImage->height);
 
       memset( padfPixelSum, 0, sizeof(double)*bandCount );
 
@@ -638,45 +591,26 @@ msAverageRasterResampler( imageObj *psSrcImage, rasterBufferObj *src_rb,
 
       if( MS_RENDERER_PLUGIN(psSrcImage->format) ) {
         assert(dst_rb && src_rb);
-#ifdef USE_GD
-        if(dst_rb->type == MS_BUFFER_GD) {
-          int nResult = panCMap[(int) padfPixelSum[0]];
-          assert( !gdImageTrueColor(dst_rb->data.gd_img) );
-          if( nResult != -1 ) {
-            nSetPoints++;
-            dst_rb->data.gd_img->pixels[nDstY][nDstX] = nResult;
-          }
-        } else
-#endif
-#ifdef USE_GD
-          if( dst_rb->type == MS_BUFFER_BYTE_RGBA ) {
-#else
-          assert( dst_rb->type == MS_BUFFER_BYTE_RGBA );
-#endif
+        assert( dst_rb->type == MS_BUFFER_BYTE_RGBA );
+        assert( src_rb->type == dst_rb ->type );
 
+        nSetPoints++;
 
-            assert( src_rb->type == dst_rb ->type );
+        if( dfAlpha01 > 0 ) {
+          unsigned char red, green, blue, alpha;
 
-            nSetPoints++;
+          red   = (unsigned char)
+                  MS_MAX(0,MS_MIN(255,padfPixelSum[0]+0.5));
+          green = (unsigned char)
+                  MS_MAX(0,MS_MIN(255,padfPixelSum[1]+0.5));
+          blue  = (unsigned char)
+                  MS_MAX(0,MS_MIN(255,padfPixelSum[2]+0.5));
+          alpha = (unsigned char)
+                  MS_MAX(0,MS_MIN(255,255*dfAlpha01+0.5));
 
-            if( dfAlpha01 > 0 ) {
-              unsigned char red, green, blue, alpha;
-
-              red   = (unsigned char)
-                      MAX(0,MIN(255,padfPixelSum[0]+0.5));
-              green = (unsigned char)
-                      MAX(0,MIN(255,padfPixelSum[1]+0.5));
-              blue  = (unsigned char)
-                      MAX(0,MIN(255,padfPixelSum[2]+0.5));
-              alpha = (unsigned char)
-                      MAX(0,MIN(255,255*dfAlpha01+0.5));
-
-              RB_SET_PIXEL(dst_rb,nDstX,nDstY,
-                           red, green, blue, alpha );
-            }
-#ifdef USE_GD
-          }
-#endif
+          RB_MIX_PIXEL(dst_rb,nDstX,nDstY,
+                       red, green, blue, alpha );
+        }
       } else if( MS_RENDERER_RAWDATA(psSrcImage->format) ) {
         int band;
         int dst_off = nDstX + nDstY * psDstImage->width;
@@ -777,8 +711,10 @@ void *msInitProjTransformer( projectionObj *psSrc,
     psPTInfo->bSrcIsGeographic = MS_FALSE;
 
   if( !InvGeoTransform(padfSrcGeoTransform,
-                       psPTInfo->adfInvSrcGeoTransform) )
+                       psPTInfo->adfInvSrcGeoTransform) ) {
+    free(psPTInfo);
     return NULL;
+  }
 
   /* -------------------------------------------------------------------- */
   /*      Record destination image information.                           */
@@ -1180,10 +1116,10 @@ static int msTransformMapToSource( int nDstXSize, int nDstYSize,
       psSrcExtent->miny = psSrcExtent->maxy = y_out;
       bOutInit = 1;
     } else {
-      psSrcExtent->minx = MIN(psSrcExtent->minx, x_out);
-      psSrcExtent->maxx = MAX(psSrcExtent->maxx, x_out);
-      psSrcExtent->miny = MIN(psSrcExtent->miny, y_out);
-      psSrcExtent->maxy = MAX(psSrcExtent->maxy, y_out);
+      psSrcExtent->minx = MS_MIN(psSrcExtent->minx, x_out);
+      psSrcExtent->maxx = MS_MAX(psSrcExtent->maxx, x_out);
+      psSrcExtent->miny = MS_MIN(psSrcExtent->miny, y_out);
+      psSrcExtent->maxy = MS_MAX(psSrcExtent->maxy, y_out);
     }
   }
 
@@ -1200,10 +1136,10 @@ static int msTransformMapToSource( int nDstXSize, int nDstYSize,
     int nGrowAmountY = (int)
                        (psSrcExtent->maxy - psSrcExtent->miny)/EDGE_STEPS + 1;
 
-    psSrcExtent->minx = MAX(psSrcExtent->minx - nGrowAmountX,0);
-    psSrcExtent->miny = MAX(psSrcExtent->miny - nGrowAmountY,0);
-    psSrcExtent->maxx = MIN(psSrcExtent->maxx + nGrowAmountX,nSrcXSize);
-    psSrcExtent->maxy = MIN(psSrcExtent->maxy + nGrowAmountY,nSrcYSize);
+    psSrcExtent->minx = MS_MAX(psSrcExtent->minx - nGrowAmountX,0);
+    psSrcExtent->miny = MS_MAX(psSrcExtent->miny - nGrowAmountY,0);
+    psSrcExtent->maxx = MS_MIN(psSrcExtent->maxx + nGrowAmountX,nSrcXSize);
+    psSrcExtent->maxy = MS_MIN(psSrcExtent->maxy + nGrowAmountY,nSrcYSize);
   }
 
   return MS_TRUE;
@@ -1244,6 +1180,7 @@ int msResampleGDALToMap( mapObj *map, layerObj *layer, imageObj *image,
   int         nLoadImgXSize, nLoadImgYSize;
   double      dfOversampleRatio;
   rasterBufferObj src_rb, *psrc_rb = NULL, *mask_rb = NULL;
+  int         bAddPixelMargin = MS_TRUE;
 
 
   const char *resampleMode = CSLFetchNameValue( layer->processing,
@@ -1335,21 +1272,75 @@ int msResampleGDALToMap( mapObj *map, layerObj *layer, imageObj *image,
     sSrcExtent.maxy = nSrcYSize;
   }
 
+  /* -------------------------------------------------------------------- */
+  /*      If requesting at the raster resolution, on pixel boundaries,    */
+  /*      and no reprojection is  involved, we don't need any resampling. */
+  /*      And if they match an integral subsampling factor, no need to    */
+  /*      add pixel margin.                                               */
+  /*      This optimization helps a lot when operating with mode=tile     */
+  /*      and that the underlying raster is tiled and share the same      */
+  /*      tiling scheme as the queried tile mode.                         */
+  /* -------------------------------------------------------------------- */
+#define IS_ALMOST_INTEGER(x, eps)        (fabs((x)-(int)((x)+0.5)) < (eps))
+
+  if( adfSrcGeoTransform[1] > 0.0 && adfSrcGeoTransform[2] == 0.0 &&
+      adfSrcGeoTransform[4] == 0.0 && adfSrcGeoTransform[5] < 0.0 &&
+      IS_ALMOST_INTEGER(sSrcExtent.minx, 0.1) &&
+      IS_ALMOST_INTEGER(sSrcExtent.miny, 0.1) &&
+      IS_ALMOST_INTEGER(sSrcExtent.maxx, 0.1) &&
+      IS_ALMOST_INTEGER(sSrcExtent.maxy, 0.1) &&
+      !msProjectionsDiffer( &(map->projection), &(layer->projection)) ) 
+  {
+      double dfXFactor, dfYFactor;
+
+      sSrcExtent.minx = (int)(sSrcExtent.minx + 0.5);
+      sSrcExtent.miny = (int)(sSrcExtent.miny + 0.5);
+      sSrcExtent.maxx = (int)(sSrcExtent.maxx + 0.5);
+      sSrcExtent.maxy = (int)(sSrcExtent.maxy + 0.5);
+      
+      if( (int)(sSrcExtent.maxx - sSrcExtent.minx + 0.5) == nDstXSize &&
+          (int)(sSrcExtent.maxy - sSrcExtent.miny + 0.5) == nDstYSize )
+      {
+            if( layer->debug )
+                msDebug( "msResampleGDALToMap(): Request matching raster resolution and pixel boundaries. "
+                         "No need to do resampling/reprojection.\n" );
+            return msDrawRasterLayerGDAL( map, layer, image, rb, hDS );
+      }
+
+      dfXFactor = (sSrcExtent.maxx - sSrcExtent.minx) / nDstXSize;
+      dfYFactor = (sSrcExtent.maxy - sSrcExtent.miny) / nDstYSize;
+      if( IS_ALMOST_INTEGER(dfXFactor, 1e-5) &&
+          IS_ALMOST_INTEGER(dfYFactor, 1e-5) &&
+          IS_ALMOST_INTEGER(sSrcExtent.minx/dfXFactor, 0.1) &&
+          IS_ALMOST_INTEGER(sSrcExtent.miny/dfXFactor, 0.1) &&
+          IS_ALMOST_INTEGER(sSrcExtent.maxx/dfYFactor, 0.1) &&
+          IS_ALMOST_INTEGER(sSrcExtent.maxy/dfYFactor, 0.1) )
+      {
+          bAddPixelMargin = MS_FALSE;
+          if( layer->debug )
+            msDebug( "msResampleGDALToMap(): Request matching raster resolution "
+                     "and pixel boundaries matching an integral subsampling factor\n" );
+      }
+  }
+
   /* -------------------------------------------------------------------- */
   /*      Project desired extents out by 2 pixels, and then strip to      */
   /*      available data.                                                 */
   /* -------------------------------------------------------------------- */
   memcpy( &sOrigSrcExtent, &sSrcExtent, sizeof(sSrcExtent) );
 
-  sSrcExtent.minx = floor(sSrcExtent.minx-1.0);
-  sSrcExtent.maxx = ceil (sSrcExtent.maxx+1.0);
-  sSrcExtent.miny = floor(sSrcExtent.miny-1.0);
-  sSrcExtent.maxy = ceil (sSrcExtent.maxy+1.0);
+  if( bAddPixelMargin )
+  {
+    sSrcExtent.minx = floor(sSrcExtent.minx-1.0);
+    sSrcExtent.maxx = ceil (sSrcExtent.maxx+1.0);
+    sSrcExtent.miny = floor(sSrcExtent.miny-1.0);
+    sSrcExtent.maxy = ceil (sSrcExtent.maxy+1.0);
+  }
 
-  sSrcExtent.minx = MAX(0,sSrcExtent.minx);
-  sSrcExtent.maxx = MIN(sSrcExtent.maxx, nSrcXSize );
-  sSrcExtent.miny = MAX(sSrcExtent.miny, 0 );
-  sSrcExtent.maxy = MIN(sSrcExtent.maxy, nSrcYSize );
+  sSrcExtent.minx = MS_MAX(0,sSrcExtent.minx);
+  sSrcExtent.maxx = MS_MIN(sSrcExtent.maxx, nSrcXSize );
+  sSrcExtent.miny = MS_MAX(sSrcExtent.miny, 0 );
+  sSrcExtent.maxy = MS_MIN(sSrcExtent.maxy, nSrcYSize );
 
   if( sSrcExtent.maxx <= sSrcExtent.minx
       || sSrcExtent.maxy <= sSrcExtent.miny ) {
@@ -1387,9 +1378,9 @@ int msResampleGDALToMap( mapObj *map, layerObj *layer, imageObj *image,
   else
     sDummyMap.cellsize = dfNominalCellSize;
 
-  nLoadImgXSize = MAX(1, (int) (sSrcExtent.maxx - sSrcExtent.minx)
+  nLoadImgXSize = MS_MAX(1, (int) (sSrcExtent.maxx - sSrcExtent.minx)
                       * (dfNominalCellSize / sDummyMap.cellsize));
-  nLoadImgYSize = MAX(1, (int) (sSrcExtent.maxy - sSrcExtent.miny)
+  nLoadImgYSize = MS_MAX(1, (int) (sSrcExtent.maxy - sSrcExtent.miny)
                       * (dfNominalCellSize / sDummyMap.cellsize));
 
   /*
@@ -1453,28 +1444,12 @@ int msResampleGDALToMap( mapObj *map, layerObj *layer, imageObj *image,
   if( MS_RENDERER_PLUGIN(sDummyMap.outputformat) ) {
     assert(rb);
     msInitializeRendererVTable(sDummyMap.outputformat);
-#ifdef USE_GD
-    if( rb->type == MS_BUFFER_GD ) {
-      assert( !gdImageTrueColor( rb->data.gd_img ) );
-      sDummyMap.outputformat->transparent = MS_TRUE;
-      sDummyMap.imagecolor.red = 117;
-      sDummyMap.imagecolor.green = 17;
-      sDummyMap.imagecolor.blue = 191;
-    }
-    /* -------------------------------------------------------------------- */
-    /*      If we are working in RGB mode ensure we produce an RGBA         */
-    /*      image so the transparency can be preserved.                     */
-    /* -------------------------------------------------------------------- */
-    else
-#endif
-    {
-      assert( sDummyMap.outputformat->imagemode == MS_IMAGEMODE_RGB
-              || sDummyMap.outputformat->imagemode == MS_IMAGEMODE_RGBA );
+    assert( sDummyMap.outputformat->imagemode == MS_IMAGEMODE_RGB
+            || sDummyMap.outputformat->imagemode == MS_IMAGEMODE_RGBA );
 
-      sDummyMap.outputformat->transparent = MS_TRUE;
-      sDummyMap.outputformat->imagemode = MS_IMAGEMODE_RGBA;
-      MS_INIT_COLOR(sDummyMap.imagecolor,-1,-1,-1,255);
-    }
+    sDummyMap.outputformat->transparent = MS_TRUE;
+    sDummyMap.outputformat->imagemode = MS_IMAGEMODE_RGBA;
+    MS_INIT_COLOR(sDummyMap.imagecolor,-1,-1,-1,255);
   }
 
   /* -------------------------------------------------------------------- */
@@ -1492,9 +1467,13 @@ int msResampleGDALToMap( mapObj *map, layerObj *layer, imageObj *image,
     psrc_rb = &src_rb;
     memset( psrc_rb, 0, sizeof(rasterBufferObj) );
     if( srcImage->format->vtable->supports_pixel_buffer ) {
-      srcImage->format->vtable->getRasterBufferHandle( srcImage, psrc_rb );
+      if(UNLIKELY(MS_FAILURE == srcImage->format->vtable->getRasterBufferHandle( srcImage, psrc_rb ))) {
+        return -1;
+      }
     } else {
-      srcImage->format->vtable->initializeRasterBuffer(psrc_rb,nLoadImgXSize, nLoadImgYSize,MS_IMAGEMODE_RGBA);
+      if(UNLIKELY(MS_FAILURE == srcImage->format->vtable->initializeRasterBuffer(psrc_rb,nLoadImgXSize, nLoadImgYSize,MS_IMAGEMODE_RGBA))) {
+        return -1;
+      }
     }
   }
 
@@ -1525,29 +1504,6 @@ int msResampleGDALToMap( mapObj *map, layerObj *layer, imageObj *image,
     }
   }
 
-#ifdef USE_GD
-  /* -------------------------------------------------------------------- */
-  /*      Do we need to generate a colormap remapping, potentially        */
-  /*      allocating new colors on the destination color map?             */
-  /* -------------------------------------------------------------------- */
-  if( psrc_rb && psrc_rb->type == MS_BUFFER_GD ) {
-    int  iColor, nColorCount;
-
-    anCMap[0] = -1; /* color zero is always transparent */
-
-    nColorCount = gdImageColorsTotal( psrc_rb->data.gd_img );
-    for( iColor = 1; iColor < nColorCount; iColor++ ) {
-      anCMap[iColor] =
-        msAddColorGD( map, rb->data.gd_img, 0,
-                      gdImageRed( psrc_rb->data.gd_img, iColor ),
-                      gdImageGreen( psrc_rb->data.gd_img, iColor ),
-                      gdImageBlue( psrc_rb->data.gd_img, iColor ) );
-    }
-    for( iColor = nColorCount; iColor < 256; iColor++ )
-      anCMap[iColor] = -1;
-  }
-#endif
-
   /* -------------------------------------------------------------------- */
   /*      Setup transformations between our source image, and the         */
   /*      target map image.                                               */
